function readnextframe(s::MDFileStream, ::Val{Format.CFG}, fmasktimestep::Int)::MDFrame
    check_eof(s)
    H0 = zero(MMatrix{3,3,Float64})
    H0supplied = zero(MMatrix{3,3,Bool})
    A::Float64 = 1.0
    R::Float64 = 1.0
    system_properties = Dict{Symbol,Any}()
    particles = DataFrame()
    velocities_in_file = true
    is_extended_cfg = false
    entrycount = 6 # if regular cfg (not extended)
    transform = one(MMatrix{3,3,Float64})
    auxproperties = Vector{String}()
    natoms = 0

    try
        natoms = readnumparticles(readline(s))

        # system properties reading loop
        while !eof(s)
            lastpos = position(s)
            line = readline(s)
            if match(r"^\d", line) !== nothing
                # if the first symbol is a digit, we've reached the particles list
                seek(s, lastpos; lines_delta = -1)
                break
            elseif match(r"^A\s", line) !== nothing
                A = readlengthscale(line)
            elseif match(r"^H0", line) !== nothing
                i, j, val = readH0tag(line)
                H0[j, i] = val
                H0supplied[j, i] = true
            elseif match(r"^R\s", line) !== nothing
                R = readratescale(line)
            elseif match(r"^.NO_VELOCITY.\s*$", line) !== nothing
                !is_extended_cfg || error(".NO_VELOCITY. must be set before entry_count")
                velocities_in_file = false
            elseif match(r"^entry_count\s", line) !== nothing
                entrycount = readentrycount(line)
                # extended cfg has for each atom x, y, z and optionally vx, vy, vz
                auxcount = entrycount - 3 - 3 * velocities_in_file
                auxproperties = ["" for i in 1:auxcount]
                is_extended_cfg = true
            elseif match(r"^auxiliary", line) !== nothing
                i, name = readauxiliary(line)
                is_extended_cfg || error("not extended cfg (define entry_count before auxiliary tags)")
                1+i in 1:length(auxproperties) || error("aux property id $(i) out of bounds")
                auxproperties[1+i] == "" || error("aux property $(i) name already set")
                auxproperties[1+i] = name
            end
        end

        # reading particles
        # if not every H0 matrix cell was in the input, error
        sum(H0supplied) == 9 || error("Incomplete supercell")

        all([s != "" for s in auxproperties]) || error("not all aux properties set")

        if is_extended_cfg
            properties = ["xs", "ys", "zs"] 
            if velocities_in_file
                append!(properties, ["vx", "vy", "vz"])
            end
            append!(properties, auxproperties)

            particles = readparticles_extended(s, natoms, properties)
        else
            particles = readparticles(s, natoms)
        end

        seekend(s)
    catch(ex)
        if typeof(ex) === FileParsingError
            rethrow()
        end
        # wrap errors in FileParsingError
        throw(FileParsingError(s, sprint(showerror, ex)))
    end

    basis = transpose(H0 * transform)
    origin = zero(SVector{3,Float64})
    p = BoundaryCondition.Periodic
    bc = BoundaryConditions(p,p,p,p,p,p)
    triclinic = (basis[1,2] != 0 || basis[1,3] != 0 || basis[2,3] != 0 || basis[2,1] != 0 || basis[3,1] != 0 || basis[3,2] != 0)
    box = SimulationBox(basis, origin, bc, triclinic)

    system_properties[:A] = A
    system_properties[:R] = R

    return MDFrame(fmasktimestep, box, particles, system_properties)
end

# read table of particles from extended cfg
function readparticles_extended(s::MDFileStream, natoms::Int, properties::Vector{String})::DataFrame
	# lines with particle headers (transformed into form "mass element")
	headerlines = Vector{UInt8}()
	# lines with particle values
	valuelines = Vector{UInt8}()
	# lines starting with #
	commentlines = Vector{UInt8}()
	
	# read s until the end into buffer
    linenum = linenumber(s)
	buf = read(s)
	
    # holds the current line
	currline = Vector{UInt8}()

    # holds lines specifying the mass of particles (to be parsed later)
    masslines = Vector{UInt8}()
    # holds line numbers of lines with mass
    masslinenumbers = Vector{Int}()
    # holds lines specifying the element of particles (to be parsed later)
    elementlines = Vector{UInt8}()
    # holds line numbers of lines with element
    elementlinenumbers = Vector{Int}()
    # holds lines with particle property lists (to be parsed later)
    propertieslines = Vector{UInt8}()
    # line numbers of property lines
    propertieslinenumbers = Vector{Int}()

    newline::UInt8 = 0x0A
    creturn::UInt8 = 0x0D
    hashmark::UInt8 = 0x23
    space::UInt8 = 0x20

    for byte in buf
        push!(currline, byte)
        if byte == newline
            # lines are of 4 types:
            # 1) mass specification (1 float)
            # 2) element (1 string) - the line after a mass specification
            # 3) particle property list (>= 3 floats)
            # 4) comment (line starts with hashmark)

            # comment
            if currline[1] == hashmark
                linenum += 1
                resize!(currline, 0)
                continue
            end

            # count number of space-separated elements on the current line
            elct = 0
            for i in 1:length(currline)
                if !isspace(Char(currline[i])) && (i == 1 || currline[i-1] == space)
                    elct += 1
                end
            end

            prev_line_is_mass = !isempty(masslinenumbers) && last(masslinenumbers) == linenum - 1
            prev_line_is_element = !isempty(elementlinenumbers) && last(elementlinenumbers) == linenum - 1
            prev_line_is_properties = !isempty(propertieslinenumbers) && last(propertieslinenumbers) == linenum - 1
            
            # if its the first line or if theres a line with 1 token after a properties line
            # expect line specifying mass for the next block of particles
            if isempty(masslinenumbers) || (elct == 1 && prev_line_is_properties)
                if elct != 1
                    throw(FileParsingError(s, linenum, "expected mass line (1 float)"))
                end
                append!(masslines, currline)
                push!(masslinenumbers, linenum)
            elseif prev_line_is_mass
                # expect element line
                if elct != 1
                    throw(FileParsingError(s, linenum, "expected element line (1 string)"))
                end
                append!(elementlines, currline)
                push!(elementlinenumbers, linenum)
            else
                # expect particle properties line
                if elct != length(properties)
                    throw(FileParsingError(s, linenum, "expected $(length(properties)) values"))
                end
                append!(propertieslines, currline)
                push!(propertieslinenumbers, linenum)
            end

            resize!(currline, 0)
            linenum += 1
        end
    end

    if length(masslinenumbers) != length(elementlinenumbers)
        throw(FileParsingError(s, linenum, "expected element line"))
    end

    if length(propertieslinenumbers) != natoms
        throw(FileParsingError(s, linenum, "expected $(natoms) particles, got $(length(propertieslinenumbers))"))   
    end

    massdf = try
        readtable(masslines; header=["mass"], types=[Float64])
    catch ex
        if isa(ex, TableParsingError)
            throw(FileParsingError(s, masslinenumbers[ex.row], "expected float"))
        else
            rethrow()
        end
    end

    elementdf = readtable(elementlines; header=["element"], types=[String])

    particlesdf = try 
        readtable(propertieslines; header=properties, types=fill(Float64, length(properties)))
    catch ex
        if isa(ex, TableParsingError)
            throw(FileParsingError(s, propertieslinenumbers[ex.row], "property number $(ex.col) is not a float"))
        else
            rethrow()
        end
    end

    masses = Vector{Float64}()
    resize!(masses, natoms)
    elements = Vector{String}()
    resize!(elements, natoms)

    j = 1 # current particle block
    for i in 1:natoms
        # if this particle is in the next block, iterate j
        if j != length(masslinenumbers) && masslinenumbers[j+1] < propertieslinenumbers[i]
            j += 1
        end
        masses[i] = massdf[!, :mass][j]
        elements[i] = elementdf[!, :element][j]
    end

    particlesdf[!, :element] = elements
    particlesdf[!, :mass] = masses
    particlesdf
end

# read table of particles from cfg
function readparticles(s::MDFileStream, natoms::Int)::DataFrame
    tableheader = ["mass", "element", "xs", "ys", "zs", "vx", "vy", "vz"]
    tabletypes = [Float64, String, Float64, Float64, Float64, Float64, Float64, Float64]

    readtable(s; rows=natoms, header=tableheader, types=tabletypes)
end

function readauxiliary(line::AbstractString)
    m = match(r"^auxiliary\[(\d+)\]\s=\s(\S+)", line)
    m !== nothing || error("invalid auxiliary tag")
    parse(Int, m[1]), m[2]
end

function readentrycount(line::AbstractString)::Int
    m = match(r"^entry_count\s=\s(\d+)", line)
    m !== nothing || error("invalid entry count tag")
    parse(Int, m[1])
end

function readH0tag(line::AbstractString)
    m = match(r"H0\((\d),(\d)\)\s=\s(\S+)", line)
    m !== nothing || error("invalid H0 matrix value tag")
    parse(Int, m[1]), parse(Int, m[2]), parse(Float64, m[3])
end

function readratescale(line::AbstractString)::Float64
    m = match(r"R\s=\s(\S+)", line)
    m !== nothing || error("invalid rate scale tag")
    parse(Float64, m[1])
end

function readlengthscale(line::AbstractString)::Float64
    m = match(r"^A\s=\s(\S+)", line)
    m !== nothing || error("invalid length scale tag")
    parse(Float64, m[1])
end

function readnumparticles(line::AbstractString)::Int
    m = match(r"^Number\sof\sparticles\s=\s(\d+)\s*$", line)
    m !== nothing || error("expected Number of particles = ...")
    parse(Int, m[1])
end
