function readnextframe(s::MDFileStream, ::Val{Format.LAMMPSData}, fmasktimestep::Int)::MDFrame
    check_eof(s)

    natoms::Int = 0
    ntypes::Int = 0

    # box bounds
    xlo = -0.5
    xhi = 0.5
    ylo = -0.5
    yhi = 0.5
    zlo = -0.5
    zhi = 0.5
    # tilts
    xy = 0
    xz = 0
    yz = 0
    triclinic = false

    # type label to int type
    labeldict::Union{Nothing,Dict{String,Int}} = nothing
    # masses[i] corresponds to mass of atom type i
    masses::Union{Nothing,AbstractVector{Float64}} = nothing
    # per atom velocities
    velocities::Union{Nothing,DataFrame} = nothing

    timestep::Int = fmasktimestep
    particles = DataFrame()
    system_properties = Dict{Symbol,Any}()

    try
        # read file description
        # may contain timestep and units
        filedesc = readline(s)
        timestep = readtimestep(filedesc)

        if timestep == nothing
            timestep = fmasktimestep
        end

        units = readunits(filedesc)
        if units !== nothing
            system_properties[:units] = units
        end

        # main reading loop
        # first comes the header then the body
        isheader = true
        while !eof(s)
            line = readline(s)
            # skip blank lines
            if isblank(line)
                continue
            end
            if isbodysectionheader(line)
                isheader = false
                skipuntilcontent(s)
                if match(r"^Atom\sType\sLabels\b", line) !== nothing
                    labeldict = readtypelabels(s, ntypes)
                elseif match(r"^Masses\b", line) !== nothing
                    masses = readmasses(s, ntypes, labeldict)
                elseif match(r"^Velocities\b", line) !== nothing
                	nrow(particles) > 0 || error("Velocities section must come after Atoms section")
                    velocities = readvelocities(s, particles[!, :id])
                elseif match(r"^Atoms\b", line) !== nothing
                    particles = readatoms(s, natoms, line, labeldict)
                else
                    # unsupported section, read until next section
                    while !eof(s)
                        lastpos = position(s)
                        line = readline(s)
                        if isbodysectionheader(line)
                            # 'unread' line
                            seek(s, lastpos; lines_delta = -1)
                            break
                        end
                    end
                end
            else
                if !isheader
                    error("expected body section header")
                end

                if match(r"atoms", line) !== nothing
                    natoms = readnatoms(line)
                    if ntypes == 0 # if ntypes not defined yet, default to num of atoms
                        ntypes = natoms
                    end
                elseif match(r"atom\stypes", line) !== nothing
                    ntypes = readntypes(line)
                elseif match(r"xlo\sxhi", line) !== nothing
                    xlo, xhi = readxloxhi(line)
                elseif match(r"ylo\syhi", line) !== nothing
                    ylo, yhi = readyloyhi(line)
                elseif match(r"zlo\szhi", line) !== nothing
                    zlo, zhi = readzlozhi(line)
                elseif match(r"xy\sxz\syz", line) !== nothing
                    xy, xz, yz = readtilts(line)
                    triclinic = true
                end
            end
        end
    catch(ex)
        if typeof(ex) === FileParsingError
            rethrow()
        end
        # wrap errors in FileParsingError
        throw(FileParsingError(s, sprint(showerror, ex)))
    end

    basis = SMatrix{3,3,Float64}(xhi-xlo,xy,xz,0,yhi-ylo,yz,0,0,zhi-zlo)
    origin = SVector{3,Float64}(xlo,ylo,zlo)
    p = BoundaryCondition.Periodic
    bc = BoundaryConditions(p,p,p,p,p,p)
    box = SimulationBox(basis, origin, bc, triclinic)

    if masses !== nothing
        particles[!, :mass] = [masses[typeid] for typeid in particles[!, :type]]
    end

    if velocities !== nothing
        particles[!, :vx] = velocities[!, :vx]
        particles[!, :vy] = velocities[!, :vy]
        particles[!, :vz] = velocities[!, :vz]
    end

    MDFrame(timestep, box, particles, system_properties)
end

function readatoms(s::MDFileStream, natoms::Int, header::String, labeldict)::DataFrame
    # header may have a comment specifying the atom style
    m = match(r"#\s*(\w+)\s*", header)
    style::Union{Nothing,AbstractString} = nothing
    if m !== nothing
        style = m[1]
    end

    # atomic: id, type, x, y, z, (ix, iy, iz)?
    # full: id, mol, type, q, x, y, z, (ix, iy, iz)?
    lastpos = position(s)
    line = readline(s)
    seek(s, lastpos; lines_delta = -1)
    cols = length(split(line))
    # each style can optionally have image flags appended (ix, iy, iz)
    if style === nothing
        # try to guess
        style = if cols == 5 || cols == 8
            "atomic"
        elseif cols == 7 || cols == 10
            "full"
        end
    end

    style == "atomic" || style == "full" || error("only atomic and full styles supported")

    # If labeldict is given, must parse type column as strings
    # to convert type labels to type ids later
    atomtype_datatype = labeldict === nothing ? Int : String

    local tableheader
    local tabletypes
    if style == "atomic"
        tableheader = ["id", "type", "x", "y", "z"]
        tabletypes = [Int, atomtype_datatype, Float64, Float64, Float64]
    elseif style == "full"
        tableheader = ["id", "mol", "type", "q", "x", "y", "z"]
        tabletypes = [Int, Int, atomtype_datatype, Float64, Float64, Float64, Float64]
    end
    
    hasimages = cols == 8 || cols == 10
    if hasimages
        append!(tableheader, ["ix", "iy", "iz"])
        append!(tabletypes, [Int16, Int16, Int16])
    end

    df = readtable(s; rows=natoms, header=tableheader, types=tabletypes)
    sort!(df, [:id])
    df[!, :id] == 1:natoms || error("expected data for each atom")

    if labeldict !== nothing
        df[!, :type] = labelstotypes(df[!, :type], labeldict)
    end

    df
end

function readvelocities(s::MDFileStream, ids::Vector{Int})::DataFrame
    tableheader = ["id", "vx", "vy", "vz"]
    tabletypes = [Int, Float64, Float64, Float64]

    df = readtable(s; rows=length(ids), header=tableheader, types=tabletypes)
    sort!(df, [:id])
    df[!, :id] == ids || error("set of ids in Velocities must match set of ids in Atoms)")

    df
end

function readmasses(s::MDFileStream, ntypes::Int, labeldict)::AbstractVector{Float64}
    tableheader = ["type", "mass"]
    tabletypes = labeldict === nothing ? [Int, Float64] : [String, Float64]
    
    df = readtable(s; rows=ntypes, header=tableheader, types=tabletypes)
    if labeldict !== nothing
        df[!, :type] = labelstotypes(df[!, :type], labeldict)
    end
    
    sort!(df, [:type])
    df[!, :type] == 1:ntypes || error("types from 1 to ntypes must be assigned masses")
    df[!, :mass]
end

function labelstotypes(labels::Vector{String}, labeldict::Dict{String,Int})::Vector{Int}
    types = Vector{Int}()
    for label in labels
        if label in keys(labeldict)
            push!(types, labeldict[label])
        else
            type = tryparse(Int, label)
            if type === nothing
                error("Type label $(label) not found")
            end
            push!(types, type)
        end
    end
    types
end

function readtypelabels(s::MDFileStream, ntypes::Int)::Dict{String,Int}
    typelabels = Dict{String,Int}()

    tableheader = ["type", "label"]
    types = [Int, String]
    
    df = readtable(s; rows=ntypes, header=tableheader, types=types)

    sort!(df, [:type])
    df[!, :type] == 1:ntypes || error("types from 1 to ntypes must be assigned labels")
    length(unique(df[!, :label])) == ntypes || error("type labels must be unique")

    Dict([row.label => row.type for row in eachrow(df)])
end

function skipuntilcontent(s::MDFileStream)
    while !eof(s)
        lastpos = position(s)
        line = readline(s)
        if !isblank(line)
            seek(s, lastpos; lines_delta = -1)
            return
        end
    end
    unexpected_eof(s)
end

function isbodysectionheader(line::AbstractString)::Bool
    # a body section always starts with an uppercase letter
    # header sections dont
    match(r"^[A-Z]", line) !== nothing
end

function isblank(line::AbstractString)::Bool
    re = r"^\s*(?:#|$)"
    match(re, line) !== nothing
end

function readxloxhi(line::AbstractString)
    m = match(r"^(\S+)\s+(\S+)\s+xlo\sxhi\s*$", line)
    if m == nothing
        error("invalid xlo xhi header line")
    end
    xlo = parse(Float64, m[1])
    xhi = parse(Float64, m[2])
    xlo, xhi
end

function readyloyhi(line::AbstractString)
    m = match(r"^(\S+)\s+(\S+)\s+ylo\syhi\s*$", line)
    if m == nothing
        error("invalid ylo yhi header line")
    end
    ylo = parse(Float64, m[1])
    yhi = parse(Float64, m[2])
    ylo, yhi
end

function readzlozhi(line::AbstractString)
    m = match(r"^(\S+)\s+(\S+)\s+zlo\szhi\s*$", line)
    if m == nothing
        error("invalid zlo zhi header line")
    end
    zlo = parse(Float64, m[1])
    zhi = parse(Float64, m[2])
    zlo, zhi
end

function readtilts(line::AbstractString)
    m = match(r"^(\S+)\s+(\S+)\s+(\S+)\s+xy\sxz\syz\s*$", line)
    if m == nothing
        error("invalid xy xz yz (tilts) header line")
    end
    xy = parse(Float64, m[1])
    xz = parse(Float64, m[2])
    yz = parse(Float64, m[3])
    xy, xz, yz
end

function readntypes(line::AbstractString)::Int
    m = match(r"^(\d+)\s+atom\stypes\s*$", line)
    if m == nothing
        error("invalid number of atom types header line")
    end
    ntypes = parse(Int, m[1])
    ntypes >= 0 || error("number of atom types must be >= 0")
    ntypes
end

function readnatoms(line::AbstractString)::Int
    m = match(r"^(\d+)\s+atoms\s*$", line)
    if m == nothing
        error("invalid number of atoms header line")
    end
    natoms = parse(Int, m[1])
    natoms >= 0 || error("number of atoms must be >= 0")
    natoms
end

function readtimestep(filedesc::AbstractString)::Union{Int,Nothing}
    re = r"\btimestep = (\d+)"
    m = match(re, filedesc)
    m !== nothing ? parse(Int, m[1]) : nothing
end

function readunits(filedesc::AbstractString)::Union{AbstractString,Nothing}
    re = r"\bunits = (\w+)\b"
    m = match(re, filedesc)
    m !== nothing ? m[1] : nothing
end
