const item_names = ["UNITS", "TIME", "TIMESTEP", "NUMBER OF ATOMS", "BOX BOUNDS", "ATOMS"]
const optional_item_names = ["UNITS", "TIME"]

function readnextframe(s::MDFileStream, ::Val{Format.LAMMPSDump}, fmasktimestep::Int)::MDFrame
    check_eof(s)
    
    timestep::Int = 0
    natoms::Int = 0
    time::Union{Float64,Nothing} = nothing
    units::Union{String,Nothing} = nothing
    box::Union{SimulationBox,Nothing} = nothing
    particles::Union{DataFrame,Nothing} = nothing
    
    itemheader, itemname = readitemheader(s)
    for (i, name_i) in enumerate(item_names)
        if itemname != name_i
            if !(name_i in optional_item_names)
                throw(FileParsingError(s, linenumber(s)-1, "Expected item $name_i"))
            end
            continue
        end
        
        if itemname == "UNITS"
            units = readline(s)
        elseif itemname == "TIME"
            time = parseline(s, Float64)
        elseif itemname == "TIMESTEP"
            timestep = parseline(s, Int)
        elseif itemname == "NUMBER OF ATOMS"
            natoms = parseline(s, Int)
        elseif itemname == "BOX BOUNDS"
            box = readbox(s, itemheader)
        elseif itemname == "ATOMS"
            particles = readparticles(s, itemheader, natoms)
        end
        
        if i != length(item_names)
            itemheader, itemname = readitemheader(s)
        end
    end
    
    properties = Dict{Symbol,Any}()
    if time != nothing
        properties[:time] = time
    end
    if units != nothing
        properties[:units] = units
    end

    MDFrame(timestep, box, particles, properties)
end

function readitemheader(s::MDFileStream)::Tuple{String,String}
    check_eof(s)
    header = readline(s)
    try
        header, itemname(header)
    catch ex
        throw(FileParsingError(s, linenumber(s)-1, ex.msg))
    end
end

function itemname(itemheader::String)::String
    for name in item_names
        pref = "ITEM: " * name
        if length(itemheader) > length(pref)
            pref *= " "
        end
        if startswith(itemheader, pref)
            return name
        end
    end
    throw(ArgumentError("Unrecognized item"))
end

function parseline(s::MDFileStream, ::Type{T}) where T
    check_eof(s)
    try
        return parse(T, readline(s))
    catch ex
        throw(FileParsingError(s, linenumber(s)-1, ex.msg))
    end
end

function readbox(s::MDFileStream, itemheader::String)::SimulationBox
    pref = "ITEM: BOX BOUNDS "
    itemheader = strip(itemheader[length(pref)+1:length(itemheader)])    
    
    bcstr = ""
    triclinic = false
    if startswith(itemheader, "xy xz yz")
        triclinic = true
        bcstr = itemheader[length("xy xz yz")+2:length(itemheader)]
    else
        bcstr = itemheader
    end
    
    bc = try
        parse(BoundaryConditions, bcstr)
    catch ex
        throw(FileParsingError(s, linenumber(s)-1, ex.msg))
    end
    
    boxdf::Union{DataFrame,Nothing} = nothing
    
    if triclinic
        boxdf = readtable(s; rows=3, header=["lo", "hi", "tilt"], types=[Float64, Float64, Float64])
    else
        boxdf = readtable(s; rows=3, header=["lo", "hi"], types=[Float64, Float64])
    end
    
    basis = zeros(MMatrix{3,3,Float64})
    origin = zeros(MVector{3,Float64})
    
    # see LAMMPS Howto triclinic
    if triclinic
        xy = boxdf.tilt[1]
        xz = boxdf.tilt[2]
        yz = boxdf.tilt[3]

        xlo_bound = boxdf.lo[1]
        xhi_bound = boxdf.hi[1]
        ylo_bound = boxdf.lo[2]
        yhi_bound = boxdf.hi[2]
        zlo_bound = boxdf.lo[3]
        zhi_bound = boxdf.hi[3]

        xlo = xlo_bound - min(0, xy, xz, xy + xz)
        xhi = xhi_bound - max(0, xy, xz, xy + xz)
        ylo = ylo_bound - min(0, yz)
        yhi = yhi_bound - max(0, yz)
        zlo = zlo_bound
        zhi = zhi_bound

        basis[1,2] = boxdf.tilt[1]
        basis[1,3] = boxdf.tilt[2]
        basis[2,3] = boxdf.tilt[3]

        basis[1,1] = xhi - xlo
        basis[2,2] = yhi - ylo
        basis[3,3] = zhi - zlo

        origin[1] = xlo
        origin[2] = ylo
        origin[3] = zlo
    else
        for i in 1:3
            basis[i,i] = boxdf.hi[i] - boxdf.lo[i]
            origin[i] = boxdf.lo[i]
        end
    end
    
    try
        SimulationBox(basis, origin, bc, triclinic)
    catch ex
        throw(FileParsingError(s, linenumber(s)-3, ex.msg))
    end
end

attribute_types = Dict(
    "id" => Int,
    "mol" => Int,
    "proc" => Int,
    "type" => Int,
    "ix" => Int16,
    "iy" => Int16,
    "iz" => Int16,
    "element" => String,
)

function readparticles(s::MDFileStream, itemheader::String, natoms::Int)::DataFrame
    pref = "ITEM: ATOMS "
    itemheader = strip(itemheader[length(pref)+1:length(itemheader)])
    tableheader = map(string, split(itemheader))
    types = DataType[]
    
    for attribute in tableheader
        if attribute in keys(attribute_types)
            push!(types, attribute_types[attribute])
        else
            push!(types, Float64)
        end
    end
    
    readtable(s; rows=natoms, header=tableheader, types=types)
end

function Base.parse(::Type{BoundaryConditions}, str::AbstractString)
    s = split(str)
    if length(s) != 3
        throw(ArgumentError("Expected boundary conditions for each axis: $str"))
    end
    xlo, xhi = parse(NTuple{2,BoundaryCondition.T}, s[1])
    ylo, yhi = parse(NTuple{2,BoundaryCondition.T}, s[2])
    zlo, zhi = parse(NTuple{2,BoundaryCondition.T}, s[3])
    BoundaryConditions(xlo, xhi, ylo, yhi, zlo, zhi)
end

function Base.parse(::Type{NTuple{2,BoundaryCondition.T}}, str::AbstractString)
    if length(str) == 1
        return (parse(BoundaryCondition.T, str), parse(BoundaryCondition.T, str))
    elseif length(str) == 2
        return (parse(BoundaryCondition.T, str[1:1]), parse(BoundaryCondition.T, str[2:2]))
    else
        throw(ArgumentError("Boundary conditions axis string must be 1 or 2 characters long: $str"))
    end
end

function Base.parse(::Type{BoundaryCondition.T}, str::AbstractString)
    if str == "p"
        return BoundaryCondition.Periodic
    elseif str == "f"
        return BoundaryCondition.Fixed
    elseif str == "s"
        return BoundaryCondition.ShrinkWrapped
    elseif str == "m"
        return BoundaryCondition.MinShrinkWrapped
    else
        throw(ArgumentError("Invalid boundary condition character $str (expected one of p, f, s, m)"))
    end
end
