function writenextframe(io::IO, frame::MDFrame, ::Val{Format.LAMMPSDump})
    if :units in keys(frame.system_properties)
        writeunits(io, frame.system_properties[:units])
    end
    
    if :time in keys(frame.system_properties)
        writetime(io, frame.system_properties[:time])
    end

    writetimestep(io, timestep(frame))

    writelength(io, length(frame))

    writebox(io, box(frame))

    writeparticles(io, frame.particles)
end

function writeunits(io::IO, units::AbstractString)
    writeitem(io, "UNITS", units)
end

function writetime(io::IO, time::AbstractFloat)
    writeitem(io, "TIME", string(time))
end

function writelength(io::IO, len::Int)
    writeitem(io, "NUMBER OF ATOMS", string(len))
end

function writetimestep(io::IO, timestep::Int)
    writeitem(io, "TIMESTEP", string(timestep))
end

function writeitem(io::IO, itemname::String, item::AbstractString)
    write(io, "ITEM: " * itemname * "\n")
    write(io, item)
    write(io, "\n")
end

function writebox(io::IO, box::SimulationBox)
    b = basis(box)
    if b[2,1] != 0 || b[3,1] != 0 || b[3,2] != 0
        throw(ArgumentError("Can't write box with basis $(b) to LAMMPS dump: matrix must be triangular"))
    end
    write(io, "ITEM: BOX BOUNDS ")
    
    triclinic = istriclinic(box)

    if triclinic
        write(io, "xy xz yz ")
    end

    writebcs(io, box.bc)
    write(io, "\n")

    o = origin(box)
    xlo = o[1]
    xhi = xlo + b[1,1]
    ylo = o[2]
    yhi = ylo + b[2,2]
    zlo = o[3]
    zhi = zlo + b[3,3]
    xy = b[1,2]
    xz = b[1,3]
    yz = b[2,3]

    xlo_bound = xlo + min(0,xy,xz,xy+xz)
    xhi_bound = xhi + max(0,xy,xz,xy+xz)
    ylo_bound = ylo + min(0,yz)
    yhi_bound = yhi + max(0,yz)
    zlo_bound = zlo
    zhi_bound = zhi

    if triclinic
        @printf(io, "%.16e %.16e %.16e\n", xlo_bound, xhi_bound, xy)
        @printf(io, "%.16e %.16e %.16e\n", ylo_bound, yhi_bound, xz)
        @printf(io, "%.16e %.16e %.16e\n", zlo_bound, zhi_bound, yz)
    else
        @printf(io, "%.16e %.16e\n", xlo, xhi)
        @printf(io, "%.16e %.16e\n", ylo, yhi)
        @printf(io, "%.16e %.16e\n", zlo, zhi)
    end
end

function writebcs(io::IO, bcs::BoundaryConditions)
    writebc(io, bcs.xlo)
    writebc(io, bcs.xhi)
    write(io, " ")
    writebc(io, bcs.ylo)
    writebc(io, bcs.yhi)
    write(io, " ")
    writebc(io, bcs.zlo)
    writebc(io, bcs.zhi)
end

function writebc(io::IO, bc::BoundaryCondition.T)
    if bc == BoundaryCondition.Periodic
        write(io, "p")
    elseif bc == BoundaryCondition.Fixed
        write(io, "f")
    elseif bc == BoundaryCondition.ShrinkWrapped
        write(io, "s")
    elseif bc == BoundaryCondition.MinShrinkWrapped
        write(io, "m")
    else
        throw(ArgumentError("unsupported boundary condition"))
    end
end

function writeparticles(io::IO, particles::DataFrame)
    write(io, "ITEM: ATOMS ")
    write(io, join(propertynames(particles), " ") * "\n")

    CSV.write(io, particles; append=true, delim=' ')
end
