mutable struct MDFileStream
    const io::IO
    const filename::String
    linenumber::Int
    
    markposition::Int
    marklinenumber::Int

    MDFileStream(io::IO, filename::String) = new(io, filename, 1, 0, 0)
end

function Base.position(stream::MDFileStream)::Int
    position(stream.io)
end

# lines delta: how many lines to add to/subtract from the line number
# for example, if we seek to an earlier position and we know we ended up
# on the line before, lines_delta would be -1
function Base.seek(stream::MDFileStream, pos::Int; lines_delta::Int)
    stream.linenumber += lines_delta
    seek(stream.io, pos)
end

function Base.readline(stream::MDFileStream)::String
    if !eof(stream.io)
        stream.linenumber += 1
    end
    readline(stream.io; keep=false) # fix GZipStream defaulting keep to true
end

const chunksize = 1024 * 1024

function readlines(stream::MDFileStream, n::Int)::Tuple{Vector{UInt8}, Int}
    v = UInt8[]
    nr = readlines!(stream, v, n)
    v, nr
end

function readlines!(stream::MDFileStream, b::Vector{UInt8}, n::Int)::Int
    initial_n::Int = n
    lastpos::Int = position(stream.io)
    it = 0
    
    while !eof(stream.io) && n > 0
        l = it + 1
        resize!(b, it + chunksize)
        s = readbytes!(stream.io, (@view b[l:it + chunksize]), chunksize)::Int
        it += s
        for i in l:it
            if b[i] == 0x0A
                n -= 1
                stream.linenumber += 1
                if n == 0
                    it = i
                    break
                end
            end
        end
    end
    
    seek(stream.io, lastpos + it)
    resize!(b, it)
    if it > 0 && b[it] != 0x0A
        n -= 1
    end
    initial_n - n
end

function readuntil(stream::MDFileStream, delim::AbstractChar)
    s = readuntil(stream.io, delim)
    stream.linenumber += count('\n', s)
    if delim == '\n'
        stream.linenumber += 1
    end
    s
end

function Base.seekend(stream::MDFileStream)
    while !eof(stream.io)
        read(stream.io)
    end
end

function Base.eof(stream::MDFileStream)::Bool
    eof(stream.io)
end

# ! doesnt modify current line number of stream: stream.linenumber
function Base.read(stream::MDFileStream)::Vector{UInt8}
    read(stream.io)
end

# marks the current position of the stream (and saves the line number)
function Base.mark(stream::MDFileStream)
    stream.markposition = position(stream.io)
    stream.marklinenumber = linenumber(stream)
end

# returns the stream to the marked position
function Base.reset(stream::MDFileStream)
    seek(stream.io, stream.markposition)
    stream.linenumber = stream.marklinenumber
end

function linenumber(stream::MDFileStream)::Int
    stream.linenumber
end

function filename(stream::MDFileStream)::String
    stream.filename
end

function check_eof(s::MDFileStream)
    if eof(s)
        throw(unexpected_eof(s))
    end
end

