"""
$(TYPEDEF)

Represents a snapshot of the molecular simulation.

Particles are stored in a table (DataFrame) with rows representing particles and columns representing particle properties.  

Column names are in LAMMPS notation (see [dump command documentation](https://docs.lammps.org/dump.html)).

The particle table is sorted by the column with name `:id`, if it exists.  

Commonly used particle properties can be accessed by specific MDFrame methods. The properties are (in LAMMPS notation): `:id`, `:mol`, `:type`, `:x`, `:y`, `:z`, `:xs`, `:ys`, `:zs`, `:xu`, `:yu`, `:zu`, `:xsu`, `:ysu`, `:zsu`, `:ix`, `:iy`, `:iz`, `:vx`, `:vy`, `:vz`, `:element`, `:mass`. Any other properties are considered 'extra' and are accessed by using the method [`extra_scalars`](@ref).

Some of the methods for accessing particle properties generate the data if it's not present in the frame, for example: particle types may be generated from particle elements or masses.

# Fields:
* `timestep`: Timestep of the snapshot
* `box`: Simulation box specification
* `particles`: Table containing particle data.
* `system_properties`: Dict containing other arbitrary system properties, like 'units' and 'time' of LAMMPS Dump or 'A' and 'R' of CFG.
"""
struct MDFrame
    timestep::Int
    box::SimulationBox
    particles::DataFrame
    system_properties::Dict{Symbol,<:Any}

    @doc """
    $(TYPEDSIGNATURES)

    Construct a MDFrame.
    """
    function MDFrame(timestep::Int, box::SimulationBox, particles::DataFrame, system_properties::Dict{Symbol,<:Any})
        if :id in propertynames(particles)
            sort!(particles, [:id])
        end
        new(timestep, box, particles, system_properties)
    end
end

"""
$(TYPEDSIGNATURES)

Return the timestep of the snapshot.
"""
function timestep(frame::MDFrame)::Int
    frame.timestep
end

"""
$(TYPEDSIGNATURES)

Return the number of particles in the system.
"""
function Base.length(frame::MDFrame)::Int
    nrow(frame.particles)
end

"""
$(TYPEDSIGNATURES)

Return unscaled wrapped positions of particles in the system.

Corresponds to properties `:x`, `:y`, `:z`.
Performs conversion from other coordinate representations if needed and possible.
"""
function positions(frame::MDFrame)::Vector{SVector{3,Float64}}
    if has_positions(frame)
        _hcat3(frame.particles.x, frame.particles.y, frame.particles.z, Float64)
    elseif has_scaled_positions(frame)
        scaled_to_unscaled(box(frame), scaled_positions(frame))
    elseif has_scaled_unwrapped_positions(frame)
        spos = wrap_scaled_positions(box(frame), scaled_unwrapped_positions(frame))
        scaled_to_unscaled(box(frame), spos)
    elseif has_unwrapped_positions(frame)
        supos = unscaled_to_scaled(box(frame), unwrapped_positions(frame))
        spos = wrap_scaled_positions(box(frame), supos)
        scaled_to_unscaled(box(frame), spos)
    else
        throw(ArgumentError("Could not compute particle wrapped positions from the frame data"))
    end
end

"""
$(TYPEDSIGNATURES)

Return scaled wrapped positions of particles in the system.

Corresponds to properties `:xs`, `:ys`, `:zs`.
Performs conversion from other coordinate representations if needed and possible.
"""
function scaled_positions(frame::MDFrame)::Vector{SVector{3,Float64}}
    if has_scaled_positions(frame)
        _hcat3(frame.particles.xs, frame.particles.ys, frame.particles.zs, Float64)
    elseif has_positions(frame)
        unscaled_to_scaled(box(frame), positions(frame))
    elseif has_scaled_unwrapped_positions(frame)
        wrap_scaled_positions(box(frame), scaled_unwrapped_positions(frame))
    elseif has_unwrapped_positions(frame)
        supos = unscaled_to_scaled(box(frame), unwrapped_positions(frame))
        wrap_scaled_positions(box(frame), supos)
    else
        throw(ArgumentError("Could not compute particle scaled wrapped positions from the frame data"))
    end
end

"""
$(TYPEDSIGNATURES)

Return scaled unwrapped positions of particles in the system.

Corresponds to properties `:xsu`, `:ysu`, `:zsu`.
Performs conversion from other coordinate representations if needed and possible.
"""
function scaled_unwrapped_positions(frame::MDFrame)::Vector{SVector{3,Float64}}
    if has_scaled_unwrapped_positions(frame)
        _hcat3(frame.particles.xsu, frame.particles.ysu, frame.particles.zsu, Float64)
    elseif has_unwrapped_positions(frame)
        unscaled_to_scaled(box(frame), unwrapped_positions(frame))
    elseif has_scaled_positions(frame) && has_images(frame)
        unwrap_scaled_positions(box(frame), scaled_positions(frame), images(frame))
    elseif has_positions(frame) && has_images(frame)
        spos = unscaled_to_scaled(box(frame), positions(frame))
        unwrap_scaled_positions(box(frame), spos, images(frame))
    else
        throw(ArgumentError("Could not compute particle scaled unwrapped positions from the frame data"))
    end
end

"""
$(TYPEDSIGNATURES)

Return unscaled unwrapped positions of particles in the system.

Corresponds to properties `:xu`, `:yu`, `:zu`.
Performs conversion from other coordinate representations if needed and possible.
"""
function unwrapped_positions(frame::MDFrame)::Vector{SVector{3,Float64}}
    if has_unwrapped_positions(frame)
        _hcat3(frame.particles.xu, frame.particles.yu, frame.particles.zu, Float64)
    elseif has_scaled_unwrapped_positions(frame)
        scaled_to_unscaled(box(frame), scaled_unwrapped_positions(frame))
    elseif has_positions(frame) && has_images(frame)
        unwrap_positions(box(frame), positions(frame), images(frame))
    elseif has_scaled_positions(frame) && has_images(frame)
        supos = unwrap_scaled_positions(box(frame), scaled_positions(frame), images(frame))
        scaled_to_unscaled(box(frame), supos)
    else
        throw(ArgumentError("Could not compute particle unwrapped positions from the frame data"))
    end
end

"""
$(TYPEDSIGNATURES)

Return velocities of particles in the system.

Corresponds to properties `:vx`, `:vy`, `:vz`.
"""
function velocities(frame::MDFrame)::Vector{SVector{3,Float64}}
    cols = names(frame.particles)
    if "vx" in cols && "vy" in cols && "vz" in cols
        return _hcat3(frame.particles.vx, frame.particles.vy, frame.particles.vz, Float64)
    end
    throw(ArgumentError("No velocity data in the frame"))
end

"""
$(TYPEDSIGNATURES)

Return x component of velocities of particles in the system.

Corresponds to property `:vx`.
"""
function velocities_x(frame::MDFrame)::Vector{Float64}
    frame.particles.vx
end

"""
$(TYPEDSIGNATURES)

Return y component of velocities of particles in the system.

Corresponds to property `:vy`.
"""
function velocities_y(frame::MDFrame)::Vector{Float64}
    frame.particles.vy
end

"""
$(TYPEDSIGNATURES)

Return z component of velocities of particles in the system.

Corresponds to property `:vz`.
"""
function velocities_z(frame::MDFrame)::Vector{Float64}
    frame.particles.vz
end

"""
$(TYPEDSIGNATURES)

Return image coordinates of particles in the system.  

Corresponds to properties `:ix`, `:iy`, `:iz`.
Generates images from unwrapped coordinates if needed and possible.
"""
function images(frame::MDFrame)::Vector{SVector{3,Int16}}
    cols = names(frame.particles)
    if has_images(frame)
        return _hcat3(frame.particles.ix, frame.particles.iy, frame.particles.iz, Int16)
    else
        supos = try
            scaled_unwrapped_positions(frame)
        catch
            throw(ArgumentError("No image data in the frame and no unwrapped positions, can't compute image data"))
        end
        [Int16.(floor.(v)) for v in supos]
    end
end

"""
$(TYPEDSIGNATURES)

Return particle ids. Guaranteed to be sorted.

Corresponds to property `:id`.
"""
function id_tags(frame::MDFrame)::Vector{Int}
    frame.particles.id
end

"""
$(TYPEDSIGNATURES)

Return particle element types.

Corresponds to property `:element`.
"""
function elements(frame::MDFrame)::Vector{String}
    frame.particles.element
end

"""
$(TYPEDSIGNATURES)

Return particle masses.

Corresponds to property `:mass`.
"""
function masses(frame::MDFrame)::Vector{Float64}
    frame.particles.mass
end

"""
$(TYPEDSIGNATURES)

Return particle types.

Corresponds to property `:type`. If there's no particle type data in the frame, attempts to generate from 1) elements of particles 2) masses of particles.
"""
function particle_types(frame::MDFrame)::Vector{Int}
    if :type in propertynames(frame.particles)
        frame.particles.type
    elseif :element in propertynames(frame.particles)
        # generate from element
        elts = elements(frame)
        uelts = sort!(unique(elts))
        elt_to_type = Dict([uelts[i] => i for i in 1:length(uelts)])
        [elt_to_type[elt] for elt in elts]
    elseif :mass in propertynames(frame.particles)
        # generate from mass
        elts = masses(frame)
        uelts = sort!(unique(elts))
        elt_to_type = Dict([uelts[i] => i for i in 1:length(uelts)])
        [elt_to_type[elt] for elt in elts]
    else
        throw(ArgumentError("No particle type data in the frame and no element or mass data in the frame"))
    end
end

"""
$(TYPEDSIGNATURES)

Return ids of the molecules that the particles belong to.

Corresponds to property `:mol`.
"""
function mol_tags(frame::MDFrame)::Vector{Int}
    frame.particles.mol
end

"""
$(TYPEDSIGNATURES)

Return extra scalar properties of the particles.

Properties are considered 'extra' if none of the other MDFrame methods return that property. The properties not considered 'extra' are: `:id`, `:mol`, `:type`, `:x`, `:y`, `:z`, `:xs`, `:ys`, `:zs`, `:xu`, `:yu`, `:zu`, `:xsu`, `:ysu`, `:zsu`, `:ix`, `:iy`, `:iz`, `:vx`, `:vy`, `:vz`, `:element`, `:mass`
"""
function extra_scalars(frame::MDFrame)::Dict{Symbol, Vector}
    extra_scalars = Dict{Symbol, Vector}()
    exclude_set = Set([:id, :mol, :type, :x, :y, :z, :xs, :ys, :zs, :xu, :yu, :zu, :xsu, :ysu, :zsu, :ix, :iy, :iz, :vx, :vy, :vz, :element, :mass])
    for name in propertynames(frame.particles)
        if !(name in exclude_set)
            extra_scalars[name] = frame.particles[!, name]
        end
    end
    extra_scalars
end

"""
$(TYPEDSIGNATURES)
	
Add an extra scalar particle property.

`name` may not be one of `:id`, `:mol`, `:type`, `:x`, `:y`, `:z`, `:xs`, `:ys`, `:zs`, `:xu`, `:yu`, `:zu`, `:xsu`, `:ysu`, `:zsu`, `:ix`, `:iy`, `:iz`, `:vx`, `:vy`, `:vz`, `:element`, `:mass`.  
`values` must be of length `length(frame)`.  
`values[i]` corresponds to the value of the added property of particle with id `id_tags(frame)[i]`.
"""
function add_extra_scalar!(frame::MDFrame, name::Symbol, values::AbstractVector{Float64})
	exclude_set = Set([:id, :mol, :type, :x, :y, :z, :xs, :ys, :zs, :xu, :yu, :zu, :xsu, :ysu, :zsu, :ix, :iy, :iz, :vx, :vy, :vz, :element, :mass])
	if name in exclude_set
		throw(ArgumentError("Name of property $(name) is not an 'extra' property"))
	end
	if length(values) != length(frame)
		throw(ArgumentError("length of values ($(length(values))) is not equal to length of frame ($(length(frame)))"))
	end
	frame.particles[!, name] = values
end

function _hcat3(v1::AbstractVector, v2::AbstractVector, v3::AbstractVector, ::Type{T})::Vector{SVector{3,T}} where T
    vt1::Vector = v1
    vt2::Vector = v2
    vt3::Vector = v3
    n = length(v1)
    [SVector{3,T}(vt1[i], vt2[i], vt3[i]) for i in 1:n]
end

"""
$(TYPEDSIGNATURES)

Return the basis vectors of the simulation box as a matrix. Simulation cell vectors are in the columns.  
"""
function boxvectors(frame::MDFrame)::SMatrix{3,3,Float64}
    basis(box(frame))
end

"""
$(TYPEDSIGNATURES)

Return the origin of the simulation box.
"""
function boxorigin(frame::MDFrame)::SVector{3,Float64}
    origin(box(frame))
end

"""
$(TYPEDSIGNATURES)

Return periodic boundary conditions.
"""
function pbc(frame::MDFrame)::NTuple{3,Bool}
    pbc(box(frame))
end

"""
$(TYPEDSIGNATURES)

Return boundary conditions.
"""
function bc(frame::MDFrame)::BoundaryConditions
    bc(box(frame))
end

"""
$(TYPEDSIGNATURES)

Return if the simulation box of the frame is triclinic.
"""
function istriclinic(frame::MDFrame)::Bool
    istriclinic(box(frame))
end

"""
$(TYPEDSIGNATURES)

Return other system properties, such as "units" and "time" of lammps dump or "A" and "R" of CFG.
"""
function system_properties(frame::MDFrame)::Dict{Symbol,<:Any}
    frame.system_properties
end

"""
$(TYPEDSIGNATURES)

Return the simulation box of the frame.
"""
function box(frame::MDFrame)::SimulationBox
    frame.box
end

function has_positions(frame::MDFrame)
    cols = names(frame.particles)
    "x" in cols && "y" in cols && "z" in cols
end

function has_scaled_positions(frame::MDFrame)
    cols = names(frame.particles)
    "xs" in cols && "ys" in cols && "zs" in cols
end

function has_unwrapped_positions(frame::MDFrame)
    cols = names(frame.particles)
    "xu" in cols && "yu" in cols && "zu" in cols
end

function has_scaled_unwrapped_positions(frame::MDFrame)
    cols = names(frame.particles)
    "xsu" in cols && "ysu" in cols && "zsu" in cols
end

function has_images(frame::MDFrame)
    cols = names(frame.particles)
    "ix" in cols && "iy" in cols && "iz" in cols
end
