"""
Enum-module containing constants corresponding to boundary condition types.

Possible boundary conditions are the same as in LAMMPS.
"""
@enumx BoundaryCondition begin
    "Periodic boundary condition."
    Periodic
    "Fixed boundary condition."
    Fixed
    "Shrink-wrapped boundary condition."
    ShrinkWrapped
    "Shrink-wrapped with a minimum value boundary condition."
    MinShrinkWrapped
end

"""
$(TYPEDEF)

Stores boundary conditions for each face of the simulation box.

# Fields:
* `xlo`: boundary style of the lower face of the box in dimension x
* `xhi`: boundary style of the higher face of the box in dimension x
* `ylo`: boundary style of the lower face of the box in dimension y
* `yhi`: boundary style of the higher face of the box in dimension y
* `zlo`: boundary style of the lower face of the box in dimension z
* `zhi`: boundary style of the higher face of the box in dimension z
"""
struct BoundaryConditions
    xlo::BoundaryCondition.T
    xhi::BoundaryCondition.T
    ylo::BoundaryCondition.T
    yhi::BoundaryCondition.T
    zlo::BoundaryCondition.T
    zhi::BoundaryCondition.T
    
    @doc """
    $(TYPEDSIGNATURES)

    Construct boundary conditions, specifying boundary type for each face. See [`BoundaryCondition`](@ref).
    """
    function BoundaryConditions(xlo, xhi, ylo, yhi, zlo, zhi)
        checkperiodicity(xlo, xhi, "x")
        checkperiodicity(ylo, yhi, "y")
        checkperiodicity(zlo, zhi, "z")
        new(xlo, xhi, ylo, yhi, zlo, zhi)
    end
end

function checkperiodicity(lo::BoundaryCondition.T, hi::BoundaryCondition.T, axis::String)
    if (lo === BoundaryCondition.Periodic) ⊻ (hi === BoundaryCondition.Periodic)
        throw(ArgumentError("Both or none of the $axis axis boundaries may be periodic: $lo $hi"))
    end
end

"""
$(TYPEDSIGNATURES)

Return periodic boundary conditions (if the boundary style for each dimension is periodic).
"""
function pbc(bc::BoundaryConditions)::NTuple{3,Bool}
    p = BoundaryCondition.Periodic
    bc.xlo === p, bc.ylo === p, bc.zlo === p
end

"""
$(TYPEDEF)

Stores the properties of the simulation box of a particle system.

# Fields:
* `basis`: simulation box vectors stored in columns
* `origin`: simulation box origin
* `bc`: boundary conditions
* `istriclinic`: box is triclinic flag
"""
struct SimulationBox
    basis::SMatrix{3,3,Float64}
    origin::SVector{3,Float64}
    bc::BoundaryConditions
    istriclinic::Bool

    @doc """
    $(TYPEDSIGNATURES)

    Construct simulation box.
    `basis` is a 3x3 matrix with simulation box vectors in columns. `origin` is a 3-component vector.
    """
    function SimulationBox(basis::AbstractMatrix{Float64}, origin::AbstractVector{Float64}, bc::BoundaryConditions, istriclinic::Bool)
        if det(basis) == 0
            throw(ArgumentError("$basis is not a valid basis (matrix is singular)"))
        end
        new(basis, origin, bc, istriclinic)
    end
end

"""
$(TYPEDSIGNATURES)

Return if the box is triclinic.
"""
function istriclinic(box::SimulationBox)::Bool
    box.istriclinic
end

"""
$(TYPEDSIGNATURES)

Return periodic boundary conditions.
"""
function pbc(box::SimulationBox)::NTuple{3,Bool}
    pbc(box.bc)
end

"""
$(TYPEDSIGNATURES)

Return boundary conditions.
"""
function bc(box::SimulationBox)::BoundaryConditions
    box.bc
end

"""
$(TYPEDSIGNATURES)

Return simulation box vectors in columns.
"""
function basis(box::SimulationBox)::SMatrix{3,3,Float64}
    box.basis
end

"""
$(TYPEDSIGNATURES)

Return box origin.
"""
function origin(box::SimulationBox)::SVector{3,Float64}
    box.origin
end

function avec(box::SimulationBox)::SVector{3,Float64}
    box.basis[:,1]
end

function bvec(box::SimulationBox)::SVector{3,Float64}
    box.basis[:,2]
end

function cvec(box::SimulationBox)::SVector{3,Float64}
    box.basis[:,3]
end

function scaled_to_unscaled(box::SimulationBox, scaled_positions::Vector{SVector{3,Float64}})::Vector{SVector{3,Float64}}
    b = basis(box)
    o = origin(box)
    [o + b * v for v in scaled_positions]
end

function unscaled_to_scaled(box::SimulationBox, unscaled_positions::Vector{SVector{3,Float64}})::Vector{SVector{3,Float64}}
    bi = basis(box) ^ -1
    o = origin(box)
    [bi * (v - o) for v in unscaled_positions]
end

function wrap_scaled_positions(box::SimulationBox, scaled_unwrapped_positions::Vector{SVector{3,Float64}})::Vector{SVector{3,Float64}}
    [SVector{3,Float64}(v[1] - floor(v[1]), v[2] - floor(v[2]), v[3] - floor(v[3])) for v in scaled_unwrapped_positions]
end

function unwrap_scaled_positions(box::SimulationBox, scaled_positions::Vector{SVector{3,Float64}}, images::Vector{SVector{3,Int16}})::Vector{SVector{3,Float64}}
    scaled_positions + images
end

function unwrap_positions(box::SimulationBox, positions::Vector{SVector{3,Float64}}, images::Vector{SVector{3,Int16}})::Vector{SVector{3,Float64}}
    @assert length(positions) == length(images)
    [positions[i] + box.basis * images[i] for i in 1:length(positions)]
end
