@testset "Coordinate conversions" begin

function scaled_wrapped_coords_dist(a::SVector{3,Float64}, b::SVector{3,Float64})
    # coords are within [0; 1]
    # to account for float imprecision when wrapping, coord distance is minimum of
    # 1) absolute difference
    # 2) distance of 'a' to 0 + distance of 'b' to 1
    # 3) distance of 'a' to 1 + distance of 'b' to 0
    SVector{3,Float64}([min(abs(a[i]-b[i]), abs(a[i])+abs(1-b[i]), abs(1-a[i])+abs(b[i])) for i in 1:3])
end

function test_coordinate_conversions(pos_filename, spos_filename, upos_filename, supos_filename)
    # Each file has particle positions in one representation
    # unscaled wrapped + images
    postraj = ReadTrajectory(pos_filename, Format.LAMMPSDump)
    # scaled wrapped + images
    spostraj = ReadTrajectory(spos_filename, Format.LAMMPSDump)
    # unscaled unwrapped
    upostraj = ReadTrajectory(upos_filename, Format.LAMMPSDump)
    # scaled unwrapped
    supostraj = ReadTrajectory(supos_filename, Format.LAMMPSDump)

    i = 1
    while !done(postraj)
        pos = read(postraj)
        spos = read(spostraj)
        upos = read(upostraj)
        supos = read(supostraj)

        frames = [pos, spos, upos, supos] 
        # MDFrame methods getting the positions in the corresponding representation
        getpositions = [positions, scaled_positions, unwrapped_positions, scaled_unwrapped_positions]

        # Test that we can convert between all possible pairs of representations
        for i in 1:4
            for j in 1:4
                # test convertation from (j) to (i)
                @testset "$(i) $(j)" begin
                positions1 = getpositions[i](frames[i])
                positions2 = getpositions[i](frames[j])
                if (i == 1 || i == 2) && (j == 3 || j == 4)
                    # testing conversion from unwrapped representations to wrapped
                    # when wrapping, account for floating point imprecision when measuring error
                    # convert to scaled first if unscaled
                    if i == 1
                        b = boxvectors(frames[i])
                        o = boxorigin(frames[i])
                        unscale(v) = b^-1 * (v - o)
                        positions1 = unscale.(positions1)
                        positions2 = unscale.(positions2)
                    end
                    function isapprox_sw(a, b)
                        isapprox(scaled_wrapped_coords_dist(a, b), [0,0,0], atol=0.0001)
                    end
                    @test all(isapprox_sw(p1,p2) for (p1, p2) in zip(positions1, positions2))
                else
                    @test all(isapprox(p1, p2, atol=0.0001) for (p1, p2) in zip(positions1, positions2))
                end

                end
            end
        end
    end
    @test done(spostraj)
    @test done(upostraj)
    @test done(supostraj)
end

@testset "Coordinate conversions LJ fluid (orthogonal box)" begin
    test_coordinate_conversions("./data/lj.T1.d0.6.0.c.dump", "./data/lj.T1.d0.6.0.cs.dump", "./data/lj.T1.d0.6.0.cu.dump", "./data/lj.T1.d0.6.0.csu.dump")
end

@testset "Coordinate conversions melt.dump (triclinic box)" begin
    test_coordinate_conversions("./data/melt.c.dump", "./data/melt.cs.dump", "./data/melt.cu.dump", "./data/melt.csu.dump")
end

end