@testset "Reading LAMMPS Dump" begin

function test_lj_trajectory(traj::ReadTrajectory)
    frame = read(traj)

    @test done(traj)

    @test_throws ArgumentError read(traj)

    @test timestep(frame) == 0

    k = 1.1856311014966876e+01
    @test isapprox(boxvectors(frame), [k 0 0; 0 k 0; 0 0 k])
    @test isapprox(boxorigin(frame), [0, 0, 0])
    @test pbc(frame) == (true, true, true)
    @test !istriclinic(frame)

    @test length(frame) == 1000

    @test id_tags(frame) == 1:length(frame)

    upos = unwrapped_positions(frame)
    @test length(upos) == length(frame)
    @test isapprox(upos[1], [-0.5828822122, 1.453042578, 0.3864926733])
    @test isapprox(upos[724], [3.641882384, 1.566770949, 8.512169418])
    @test isapprox(upos[1000], [12.53139516, 10.81468736, 10.80914051])

    es = extra_scalars(frame)
    @test length(es) == 2

    @test length(es[:c_epot]) == length(frame)
    @test isapprox(es[:c_epot][1], -3.720796954)
    @test isapprox(es[:c_epot][724], -3.489775069)
    @test isapprox(es[:c_epot][1000], -3.209647258)

    @test length(es[:c_ekin]) == length(frame)
    @test isapprox(es[:c_ekin][1], 0.3406362693)
    @test isapprox(es[:c_ekin][724], 1.154429353)
    @test isapprox(es[:c_ekin][1000], 0.4472413394)
end

@testset "Read single frame lammps dump file (lj.T1.d0.6.0.dump)" begin
    traj = ReadTrajectory("./data/lj.T1.d0.6.0.dump", Format.LAMMPSDump)
    test_lj_trajectory(traj)
end

@testset "Read gzipped single frame lammps dump file (lj.T1.d0.6.0.dump)" begin
    traj = ReadTrajectory("./data/lj.T1.d0.6.0.dump.gz", Format.LAMMPSDump)
    test_lj_trajectory(traj)
end

function test_2d_logo_trajectory(traj::ReadTrajectory)
    for i in 0:200:2400
        frame = read(traj)

        @test timestep(frame) == i
        @test length(frame) == 434

        ids = id_tags(frame)
        types = particle_types(frame)
        pos = positions(frame)
        vel = velocities(frame)
        sp = system_properties(frame)

        @test ids == 1:length(frame)
        @test length(types) == length(frame)
        @test length(pos) == length(frame)
        @test length(vel) == length(frame)
        @test isempty(extra_scalars(frame))

        @test isapprox(boxvectors(frame), [62 0 0; 0 14 0; 0 0 2])
        @test isapprox(boxorigin(frame), [0, 0, -1])
        @test pbc(frame) == (true, true, true)
        @test !istriclinic(frame)
        
        if i == 0
            @test sp[:units] == "lj"
            @test sp[:time] == 0
            @test types[1] == 1
            @test isapprox(pos[1], [0.5, 0.5, 0])
            @test isapprox(vel[1], [0.903631, 1.27504, 0])
        elseif i == 400
            @test sp[:time] == 2
            @test types[360] == 2
            @test isapprox(pos[360], [48.281, 11.2277, 0])
            @test isapprox(vel[360], [-1.2775, 0.570669, 0])
        elseif i == 1000
            @test sp[:time] == 5
            @test types[434] == 1
            @test isapprox(pos[434], [1.29844, 0.445933, 0])
            @test isapprox(vel[434], [1.51751, -0.111898, 0])
        elseif i == 2400
            @test sp[:time] == 12
            @test types[301] == 2
            @test isapprox(pos[301], [52.7384, 13.6657, 0])
            @test isapprox(vel[301], [0.0634789, -0.0333974, 0])
        end
    end
    @test done(traj)

    @test_throws ArgumentError read(traj)
end

@testset "Read single file trajectory (2d_logo_t200.dump)" begin
    traj = ReadTrajectory("./data/2d_logo_t200.dump", Format.LAMMPSDump)
    test_2d_logo_trajectory(traj)
end

@testset "Read single gzipped file trajectory (2d_logo_t200.dump.gz)" begin
    traj = ReadTrajectory("./data/2d_logo_t200.dump.gz", Format.LAMMPSDump)
    test_2d_logo_trajectory(traj)
end

@testset "Read multiple file trajectory (2d_logo_t200/2d_logo_*.dump)" begin
    traj = ReadTrajectory("./data/2d_logo_t200/2d_logo_*.dump", Format.LAMMPSDump)
    test_2d_logo_trajectory(traj)
end

@testset "Read multiple gzipped file trajectory (2d_logo_t200_gz/2d_logo_*.dump.gz)" begin
    traj = ReadTrajectory("./data/2d_logo_t200_gz/2d_logo_*.dump.gz", Format.LAMMPSDump)
    test_2d_logo_trajectory(traj)
end

end # testset Reading LAMMPS Dump