function Base.isequal(a::BoundaryConditions, b::BoundaryConditions)::Bool
    eq = a.xlo == b.xlo
    eq &= a.xhi == b.xhi
    eq &= a.ylo == b.ylo
    eq &= a.yhi == b.yhi
    eq &= a.zlo == b.zlo
    eq &= a.zhi == b.zhi
    eq
end

function Base.isapprox(a::SimulationBox, b::SimulationBox)::Bool
    isapprox(basis(a), basis(b)) && isapprox(origin(a), origin(b)) && isequal(bc(a), bc(b)) && istriclinic(a) == istriclinic(b)
end

function Base.isapprox(a::MDFrame, b::MDFrame)::Bool
    eq = isapprox(box(a), box(b))
    eq &= timestep(a) == timestep(b)
    # norm x -> NaN to force component-wise comparison of columns
    eq &= isapprox(a.particles, b.particles, atol=0.0001, norm=x->NaN)
    keys(a.system_properties) == keys(b.system_properties) || return false
    for key in keys(a.system_properties)
        if typeof(a.system_properties[key]) <: AbstractFloat
            eq &= isapprox(a.system_properties[key], b.system_properties[key])
        else
            eq &= a.system_properties[key] == b.system_properties[key]
        end
    end
    eq
end

function test_export(in_filename::AbstractString, out_filename::AbstractString)
    out_traj_filenames = Vector{String}()
    try
        firstframe = read(ReadTrajectory(in_filename, Format.LAMMPSDump))
        particle_properties = propertynames(firstframe.particles)

        intraj = ReadTrajectory(in_filename, Format.LAMMPSDump)
        outtraj = WriteTrajectory(out_filename, Format.LAMMPSDump, particle_properties)

        # import intraj and export to outtraj
        while !done(intraj)
            frame = read(intraj)
            write(outtraj, frame)
        end
        close(outtraj)

        # compare initial and written trajectories
        intraj1 = ReadTrajectory(in_filename, Format.LAMMPSDump)
        intraj2 = ReadTrajectory(out_filename, Format.LAMMPSDump)

        out_traj_filenames = copy(intraj2.filepaths)

        while !done(intraj1)
            @test isapprox(read(intraj1), read(intraj2))
        end

        @test done(intraj2)
    finally
        # remove out trajectory temp files
        rm.(out_traj_filenames)
    end
end

@testset "Writing LAMMPS Dump" begin

@testset "Simple single frame" begin
    out_filename = "./data/dump.12.temp"
    try
        # build a frame

        xlo = BoundaryCondition.Periodic
        xhi = BoundaryCondition.Periodic
        ylo = BoundaryCondition.Fixed
        yhi = BoundaryCondition.ShrinkWrapped
        zlo = BoundaryCondition.MinShrinkWrapped
        zhi = BoundaryCondition.Fixed
        
        fbcd = BoundaryConditions(xlo, xhi, ylo, yhi, zlo, zhi)
        fboxvectors = MMatrix{3,3,Float64}([2.345 1.2 1.5; 0 2.5 1; 0 0 4])
        fboxorigin = SVector{3,Float64}(1.2345, -2.3, 25.5998)

        fbox = SimulationBox(fboxvectors, fboxorigin, fbcd, true)
        ftimestep = 12

        fid = [2, 3, 1]
        fxu = [100.12345, 95.2, -0.003]
        fzu = [-20.0, 23.4, 50.6]

        fparticles = DataFrame(:id => fid, :xu => fxu, :zu => fzu)
        fsprop = Dict(:time => 2.3, :units => "lj")

        frame = MDFrame(ftimestep, fbox, fparticles, fsprop)

        # write
        wtraj = WriteTrajectory("./data/dump.*.temp", Format.LAMMPSDump, [:id, :xu, :zu])
        write(wtraj, frame)
        close(wtraj)

        # read & compare
        rtraj = ReadTrajectory("./data/dump.12.temp", Format.LAMMPSDump)
        frame_read = read(rtraj)

        @test isapprox(frame, frame_read)
    finally
        if isfile(out_filename)
            rm(out_filename)
        end
    end
end

@testset "Writing single frame" begin
    test_export("./data/lj.T1.d0.6.0.dump", "./data/dump.temp")
end

@testset "Writing single frame in gzip format" begin
    test_export("./data/lj.T1.d0.6.0.dump", "./data/dump.temp.gz")
end

@testset "Writing single frame with property computation" begin
    out_filename = "./data/temp.dump.2"
    try
        # read a frame with only unwrapped positions
        frame = read(ReadTrajectory("./data/lj.T1.d0.6.0.dump", Format.LAMMPSDump))
        # write every representation + images
        outprops = [:id, :x, :y, :z, :xs, :ys, :zs, :xu, :yu, :zu, :xsu, :ysu, :zsu, :ix, :iy, :iz]
        wtraj = WriteTrajectory(out_filename, Format.LAMMPSDump, outprops)
        write(wtraj, frame)
        close(wtraj)

        # compare to frame written by lammps
        frame_cmp = read(ReadTrajectory("./data/lj.T1.d0.6.0.c-all.dump", Format.LAMMPSDump))
        frame_written = read(ReadTrajectory(out_filename, Format.LAMMPSDump))

        @test isapprox(frame_cmp, frame_written)
    finally
        if isfile(out_filename)
            rm(out_filename)
        end
    end
end

@testset "Writing multiple frames in one file" begin
    test_export("./data/2d_logo_t200.dump", "./data/dump.temp")
end

@testset "Writing multiple frames in one file in gzip format" begin
    test_export("./data/2d_logo_t200.dump", "./data/dump.temp.gz")
end

@testset "Writing multiple frames in multiple files" begin
    test_export("./data/2d_logo_t200.dump", "./data/dump.*.temp")
end

@testset "Writing multiple frames in multiple files in gzip format" begin
    test_export("./data/2d_logo_t200.dump", "./data/dump.*.temp.gz")
end

end # testset Writing LAMMPS Dump